{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "8d299959-9d21-4b65-9775-a2e9845ffd5c",
   "metadata": {},
   "source": [
    "### WGAN与原始版本GAN和DCGAN差异较大，请仔细阅读文档并补充完整下面的代码。在需要补充的部分已经标注#TODO并附上相应的内容提示。"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "46231246-6adc-471e-a262-d9c6d0d3a48c",
   "metadata": {},
   "outputs": [],
   "source": [
    "# 导入必要的库\n",
    "import torch\n",
    "import torch.nn as nn\n",
    "import numpy as np\n",
    "from torch.utils.data import DataLoader\n",
    "import torchvision\n",
    "from torchvision import datasets, transforms\n",
    "from torchvision.utils import save_image\n",
    "import os\n",
    "from torch.utils.tensorboard import SummaryWriter  # TensorBoard"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "1df73c15-d470-4588-a9ea-a14b3d91ae4d",
   "metadata": {},
   "source": [
    "#### 根据文档和提示，补充完整WGAN的生成器Generator和判别器Discriminator代码：\n",
    "注意：\n",
    "1. WGAN 生成器的设计与 DCGAN 类似，但会使用 Tanh 激活函数将图像的像素值限制在 [-1, 1] 的范围内。\n",
    "2. 在 WGAN 中，判别器输出的是一个实数，而不是概率，因此不使用 sigmoid 激活函数。"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "231ecf9e-81d9-48b2-9d9c-1924ee0a6253",
   "metadata": {},
   "outputs": [],
   "source": [
    "# =============================== 生成器（Generator） ===============================\n",
    "class Generator(nn.Module):\n",
    "    def __init__(self, input_dim):\n",
    "        super(Generator, self).__init__()\n",
    "\n",
    "        # 1. 输入层：将 100 维随机噪声从input_dim投影到 32x32（1024 维）\n",
    "        #TODO   # 线性变换fc1，将输入噪声扩展到 1024 维\n",
    "\n",
    "        self.br1 = nn.Sequential(\n",
    "            #TODO   # 批归一化，加速训练并稳定收敛\n",
    "            #TODO   # ReLU 激活函数，引入非线性\n",
    "        )\n",
    "\n",
    "        # 2. 第二层：将 1024 维数据映射到 128 * 7 * 7 维\n",
    "        #TODO   # 线性变换，将数据变换为适合卷积层的维数大小\n",
    "\n",
    "        self.br2 = nn.Sequential(\n",
    "            #TODO   # 批归一化\n",
    "            #TODO   # ReLU 激活函数\n",
    "        )\n",
    "\n",
    "        # 3. 反卷积层 1：上采样，输出 64 通道的 14×14 特征图\n",
    "        self.conv1 = nn.Sequential(\n",
    "            #TODO   # 反卷积：将 7x7 放大到 14x14\n",
    "            #TODO   # 归一化，稳定训练\n",
    "            #TODO   # ReLU 激活函数\n",
    "        )\n",
    "\n",
    "        # 4. 反卷积层 2：输出 1 通道的 28×28 图像\n",
    "        self.conv2 = nn.Sequential(\n",
    "            #TODO   # 反卷积：将 14x14 放大到 28x28\n",
    "            #TODO    # WGAN 需要使用 Tanh 激活函数，将输出范围限制在 [-1, 1]\n",
    "        )\n",
    "\n",
    "    def forward(self, x):\n",
    "        x = self.br1(self.fc1(x))  # 通过全连接层，进行 BatchNorm 和 ReLU 激活\n",
    "        x = self.br2(self.fc2(x))  # 继续通过全连接层，进行 BatchNorm 和 ReLU 激活\n",
    "        x = x.reshape(-1, 128, 7, 7)  # 变形为适合卷积输入的形状 (batch, 128, 7, 7)\n",
    "        x = self.conv1(x)  # 反卷积：上采样到 14x14\n",
    "        output = self.conv2(x)  # 反卷积：上采样到 28x28\n",
    "        return output  # 返回生成的图像\n",
    "\n",
    "# =============================== 判别器（Discriminator） ===============================\n",
    "class Discriminator(nn.Module):\n",
    "    def __init__(self):\n",
    "        super(Discriminator, self).__init__()\n",
    "\n",
    "        # 1. 第一层：输入 1 通道的 28x28 图像，输出 32 通道的特征图，然后通过MaxPool2d降采样\n",
    "        self.conv1 = nn.Sequential(\n",
    "            #TODO  # 5x5 卷积核，步长为1\n",
    "            #TODO   # LeakyReLU，negative_slope参数设置为0.1\n",
    "        )\n",
    "        self.pl1 = nn.MaxPool2d(2, stride=2)\n",
    "\n",
    "        # 2. 第二层：输入 32 通道，输出 64 通道特征, 然后通过MaxPool2d降采样\n",
    "        self.conv2 = nn.Sequential(\n",
    "            #TODO   # 5x5 卷积核，步长为1\n",
    "            #TODO  # LeakyReLU 激活函数，negative_slope参数设置为0.1\n",
    "        )\n",
    "        self.pl2 = nn.MaxPool2d(2, stride=2)\n",
    "\n",
    "        # 3. 全连接层 1：将 64x4x4 维特征图转换成 1024 维向量\n",
    "        self.fc1 = nn.Sequential(\n",
    "            #TODO   # 线性变换，将 64x4x4 映射到 1024 维\n",
    "            #TODO   # LeakyReLU 激活函数\n",
    "        )\n",
    "\n",
    "        # 4. 全连接层 2：最终输出\n",
    "        #TODO # 输出一个标量作为判别结果\n",
    "\n",
    "    def forward(self, x):\n",
    "        x = self.pl1(self.conv1(x))  # 第一层卷积，降维\n",
    "        x = self.pl2(self.conv2(x))  # 第二层卷积，降维\n",
    "        x = x.view(x.shape[0], -1)  # 展平成向量\n",
    "        x = self.fc1(x)  # 通过全连接层\n",
    "        output = self.fc2(x)  # 通过最后一层全连接层，输出标量\n",
    "        return output  # 返回判别结果"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "96dfad47-ba00-4cad-bf4d-d0782d3d0d80",
   "metadata": {},
   "source": [
    "#### 补充完整主函数。\n",
    "注意：\n",
    "1. 传统的GAN通常使用[0, 1]范围的图像作为输入，但WGAN要求图像的像素值在 [-1, 1] 范围内。此时，输入图像的像素值需要做归一化，使用 (0.5,) 作为均值 (0.5) 和 (0.5,) 作为标准差，确保每个像素的值都被调整到 [-1, 1] 之间。\n",
    "2. 与传统的GAN使用Adam优化器不同，WGAN推荐使用RMSprop优化器。\n",
    "3. 在WGAN中，通常需要在每次生成器训练之前，先训练判别器多次。这种策略有助于使判别器的训练更加稳定。"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "053bc8c0-fc04-45f0-9a44-0213f61b9aab",
   "metadata": {},
   "outputs": [],
   "source": [
    "# =============================== 主函数 ===============================\n",
    "def main():\n",
    "    # 设备配置\n",
    "    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')\n",
    "\n",
    "    # 设定超参数\n",
    "    input_dim = 100\n",
    "    batch_size = 128\n",
    "    num_epoch = 30\n",
    "    clip_value = 0.01   # 判别器权重裁剪范围，确保满足 Lipschitz 条件\n",
    "\n",
    "    # 加载 MNIST 数据集\n",
    "    train_dataset = datasets.MNIST(root=\"./data/\", train=True, transform=transforms.Compose([\n",
    "        transforms.ToTensor(),\n",
    "        transforms.Normalize((0.5,), (0.5,))\n",
    "    ]), download=True)\n",
    "    train_loader = DataLoader(dataset=train_dataset, batch_size=batch_size, shuffle=True)\n",
    "\n",
    "    # 创建生成器G和判别器D，并移动到 GPU（如果可用）\n",
    "    # TODO\n",
    "    # TODO\n",
    "    \n",
    "    # 定义优化器optim_G和optim_D：使用RMSprop，学习率设置为0.00005\n",
    "    # TODO\n",
    "    # TODO\n",
    "\n",
    "    # 初始化 TensorBoard\n",
    "    writer = SummaryWriter(log_dir='./logs/experiment_wgan')\n",
    "\n",
    "    # 开始训练\n",
    "    for epoch in range(num_epoch):\n",
    "        total_loss_D, total_loss_G = 0, 0\n",
    "        for i, (real_images, _) in enumerate(train_loader):\n",
    "            # 判别器训练5次，然后训练生成器1次。提示：for循环，记得修改total_loss_D和total_loss_G的值\n",
    "            # TODO  # 判别器训练 5 次\n",
    "\n",
    "            # TODO  # 生成器训练 1 次\n",
    "\n",
    "            # 每 100 步打印一次损失\n",
    "            if (i + 1) % 100 == 0 or (i + 1) == len(train_loader):\n",
    "                print(f'Epoch {epoch:02d} | Step {i + 1:04d} / {len(train_loader)} | Loss_D {total_loss_D / (i + 1):.4f} | Loss_G {total_loss_G / (i + 1):.4f}')\n",
    "\n",
    "        # 记录损失到 TensorBoard\n",
    "        writer.add_scalar('WGAN/Loss/Discriminator', total_loss_D / len(train_loader), epoch)\n",
    "        writer.add_scalar('WGAN/Loss/Generator', total_loss_G / len(train_loader), epoch)\n",
    "\n",
    "        # 生成并保存示例图像\n",
    "        with torch.no_grad():\n",
    "            noise = torch.randn(64, input_dim, device=device)\n",
    "            fake_images = G(noise)\n",
    "\n",
    "            # 记录生成的图像到 TensorBoard\n",
    "            img_grid = torchvision.utils.make_grid(fake_images, normalize=True)\n",
    "            writer.add_image('Generated Images', img_grid, epoch)\n",
    "\n",
    "    writer.close()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "9883ad9f-7b8d-4cf8-a622-99936a3abc82",
   "metadata": {},
   "source": [
    "#### 根据文档中描述的GAN的损失函数和二元交叉熵损失相关内容，补充完善Discriminator和Generator的训练过程：\n",
    "注意：\n",
    "1. 判别器需要进行权重裁剪操作；\n",
    "2. 生成器和判别器的损失函数与GAN和DCGAN不同。"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "7f748fae-5147-4a63-8dc7-e6da56479a46",
   "metadata": {},
   "outputs": [],
   "source": [
    "# =============================== 训练判别器 ===============================\n",
    "def train_discriminator(real_images, D, G, optim_D, clip_value, batch_size, input_dim, device):\n",
    "    '''训练判别器'''\n",
    "    real_images = real_images.to(device)\n",
    "    real_output = D(real_images)\n",
    "\n",
    "    noise = torch.randn(batch_size, input_dim, device=device)\n",
    "    fake_images = G(noise).detach()\n",
    "    fake_output = D(fake_images)\n",
    "\n",
    "    #TODO  # 计算判别器的损失函数\n",
    "\n",
    "    optim_D.zero_grad()\n",
    "    loss_D.backward()\n",
    "    optim_D.step()\n",
    "\n",
    "    # 对判别器参数进行裁剪\n",
    "    for p in D.parameters():\n",
    "        p.data.clamp_(-clip_value, clip_value)\n",
    "\n",
    "    return loss_D.item()\n",
    "\n",
    "# =============================== 训练生成器 ===============================\n",
    "def train_generator(D, G, optim_G, batch_size, input_dim, device):\n",
    "    '''训练生成器'''\n",
    "    noise = torch.randn(batch_size, input_dim, device=device)\n",
    "    fake_images = G(noise)\n",
    "    fake_output = D(fake_images)\n",
    "\n",
    "    #TODO  # 计算生成器的损失函数\n",
    "\n",
    "    optim_G.zero_grad()\n",
    "    loss_G.backward()\n",
    "    optim_G.step()\n",
    "\n",
    "    return loss_G.item()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "4aa0f1be-01e8-4ff6-8bf3-5008ce586f57",
   "metadata": {},
   "outputs": [],
   "source": [
    "if __name__ == '__main__':\n",
    "    main()"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.14"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
